@import url('https://fonts.googleapis.com/css2?family=Inter:wght@100..600&display=swap');

:root {
    --md-primary-fg-color:          #f2a8ff;
    --md-primary-fg-color--light:   #f2a8ff;
    --md-primary-fg-color--dark:    #f2a8ff;
    --md-text-font:                 "Inter";
}
[data-md-color-scheme="default"] {
    --md-primary-fg-color:        #000000;
    --md-typeset-a-color:         #9c50c2;
    --md-accent-fg-color:         #c57fed;
}
[data-md-color-scheme="slate"] {
    --md-primary-fg-color:        #000000;
    --md-typeset-a-color:         #ca77d8;
    --md-accent-fg-color:         #f2a8ff;
}

.md-sidebar__scrollwrap:focus-within, .md-sidebar__scrollwrap:hover {
    scrollbar-color: var(--md-default-fg-color--lighter) #0000;
}
