Brainstorm a list of potentially useful text retrieval tasks.

Here are a few examples for your reference:
 - Provided a scientific claim as query, retrieve documents that help verify or refute the claim.
 - Search for documents that answers a FAQ-style query on children's nutrition.

Please adhere to the following guidelines:
 - Specify what the query is, and what the desired documents are.
 - Each retrieval task should cover a wide range of queries, and should not be too specific.

Your output should always be a python list of strings only, with about 20 elements, and each element corresponds to a distinct retrieval task in one sentence. Do not explain yourself or output anything else. Be creative!
