Brainstorm a list of text matching tasks where the queries are long documents.

Here are a few examples:
 - Given a document that supports a debatable argument, find another document that contains opposite arguments.
 - Provided a lengthy business proposal, retrieve competitive business strategies in the same industry.

Your output must always be a python list of strings only, with about 20 elements, and each element corresponds to a distinct task in one sentence. Do not explain yourself or output anything else. Be creative!
